;;;                  MASSACHUSETTS INSTITUTE OF TECHNOLOGY
;;;        Department of Electrical Engineering and Computer Science
;;;        6.001---Structure and Interpretation of Computer Programs
;;;                           Fall Semester, 1997
;;;                              Problem Set 7
;;;
;;;                     Code file CLASSES-EXPANDED.SCM

;;; -----------------------------------------------------------------------

;;An Object is represented as a message-passing procedure.

;;A Class is represented as a thunk which returns a message-passing
;;MAKE-OBJECT procedure which responds to a 'NEW or 'NEW-PART message by
;;returning an instance-object of the class.

;;;(It would be simpler to define the Class directly to be the MAKE-OBJECT
;;;procedure.  However, doing this entangles the nonempty frame created by a
;;;call to the MAKE-OBJECT procedure in the environment of the
;;;instance-object created by the call.  This frame is unnecessary, since the
;;;instance-object will keep the frame data in its slots anyway.  Forcing the
;;;frame to stay around can seriously interfere with garbage-collection of
;;;obsolete objects.  Our class-as-thunk ensures that the call-frame
;;;entangled in the instance-object will always be empty.)


(define (named-class)
  (let ((self (quote uninitialized-self)) (name (quote uninitialized-name)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (eq? class-name (quote named-class))))
                  (name (lambda () name))
                  (say
                   (lambda (list-of-stuff)
                     (newline)
                     (if (pair? list-of-stuff)
                         (display-message (list name "says nothing."))
                         (display-message
                          (cons name (cons "says" list-of-stuff))))
                     (quote nuf-said))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote name)) name)
                      ((eq? message (quote say)) say)
                      (else (get-method message))))))
           (constructor
            (lambda (initial-name)
              (set! name initial-name)
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value


(define (mobile-class)
  (let ((self (quote uninitialized-self))
        (named-part (quote uninitialized-part))
        (location (quote uninitialized-location)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote mobile-class))
                         (ask named-part (quote instance-of?) class-name))))
                  (place (lambda () location))
                  (go-to
                   (lambda (new-place)
                     (ask location (quote del-thing) self)
                     (set! location new-place)
                     (ask location (quote add-thing) self)
                     new-place)))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote place)) place)
                      ((eq? message (quote go-to)) go-to)
                      (else (get-method message named-part))))))
           (constructor
            (lambda (initial-name initial-location)
              (set! named-part (new-part named-class self initial-name))
              (set! location initial-location)
              (ask location (quote add-thing) self)
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value

;Value: places


(define (place-class)
  (let ((self (quote uninitialized-self))
        (named-part (quote uninitialized-part))
        (things (quote ()))
        (neighbor-map (quote ())))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote place-class))
                         (ask named-part (quote instance-of?) class-name))))
                  (things (lambda () things))
                  (neighbors (lambda () (map cadr neighbor-map)))
                  (exits (lambda () (map car neighbor-map)))
                  (neighbor-towards
                   (lambda (direction)
                     (let ((neighbor (assq direction neighbor-map)))
                       (and neighbor
                            (cadr neighbor)))))
                  (add-neighbor
                   (lambda (direction new-neighbor)
                     (if (ask self (quote neighbor-towards) direction)
                         (ask
                          self
                          (quote say)
                          (list (ask self (quote name))
                                "already has a neighbor to the"
                                direction))
                         (begin
                          (set!
                           neighbor-map
                           (cons (list direction new-neighbor) neighbor-map))
                          (quote added)))))
                  (have-thing? (lambda (thing) (memq thing things)))
                  (add-thing
                   (lambda (new-thing)
                     (or (ask self (quote have-thing?) new-thing)
                         (begin (set! things (cons new-thing things)) #t))))
                  (del-thing
                   (lambda (thing)
                     (begin
                      (if (ask self (quote have-thing?) thing)
                          (set! things (delq thing things))
                          (ask
                           self
                           (quote say)
                           (list (ask thing (quote name))
                                 "is not at"
                                 (ask self (quote name)))))
                      #f))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote things)) things)
                      ((eq? message (quote neighbors)) neighbors)
                      ((eq? message (quote exits)) exits)
                      ((eq? message (quote neighbor-towards)) neighbor-towards)
                      ((eq? message (quote add-neighbor)) add-neighbor)
                      ((eq? message (quote have-thing?)) have-thing?)
                      ((eq? message (quote add-thing)) add-thing)
                      ((eq? message (quote del-thing)) del-thing)
                      (else (get-method message named-part))))))
           (constructor
            (lambda (initial-name)
              (set! places (cons self places))
              (set! named-part (new-part named-class self initial-name))
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value

;Value: people


(define (person-class)
  (let ((self (quote uninitialized-self))
        (mobile-part (quote uninitialized-part))
        (possessions (quote ()))
        (restlessness (quote uninitialized-restlessness)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote person-class))
                         (ask mobile-part (quote instance-of?) class-name))))
                  (possessions (lambda () possessions))
                  (describe-possessions
                   (lambda ()
                     (ask
                      self
                      (quote say)
                      (cons
                       "I have"
                       (if (null? possessions)
                           (quote ("nothing."))
                           (map (lambda (p) (ask p (quote name)))
                                possessions))))))
                  (say
                   (lambda (list-of-stuff)
                     (newline)
                     (display-message
                      (append
                       (list "At"
                             (ask (ask self (quote place)) (quote name))
                             (ask self (quote name))
                             "says --")
                       (if (null? list-of-stuff)
                           (quote ("Oh, nevermind."))
                           list-of-stuff)))
                     (quote said)))
                  (have-fit
                   (lambda ()
                     (ask self (quote say) (quote ("Yaaaah! I am upset!")))))
                  (have-big-fit
                   (lambda ()
                     (ask
                      self
                      (quote say)
                      (quote
                       ("
                   Grendel, now be still,
                   I kill'd not thee with half so good a will.")))
                     (quote game-over)))
                  (look-around
                   (lambda ()
                     (let ((other-things
                            (map
                             (lambda (thing)
                               (ask thing (quote name)))
                             (delq
                              self
                              (ask (ask self (quote place)) (quote things))))))
                       (ask
                        self
                        (quote say)
                        (cons
                         "I see"
                         (if (null? other-things)
                             (quote ("nothing."))
                             other-things)))
                       other-things)))
                  (other-people
                   (lambda ()
                     (filter
                      (lambda (object)
                        (if (eq? object self)
                            #f
                            (ask object
                                 (quote instance-of?)
                                 (quote person-class))))
                      (ask (ask self (quote place)) (quote things)))))
                  (take
                   (lambda (thing)
                     (cond ((memq thing possessions)
                            (ask
                             self
                             (quote say)
                             (list "I already have" (ask thing (quote name)))))
                           ((and (let ((things-at-place
                                        (ask (ask self (quote place))
                                             (quote things))))
                                   (memq thing things-at-place))
                                 (ask thing
                                      (quote instance-of?)
                                      (quote thing-class)))
                            (if (ask thing (quote owned?))
                                (let ((owner (ask thing (quote owner))))
                                  (ask owner (quote lose) thing)
                                  (ask owner (quote have-fit)))
                                (quote unowned))
                            (ask thing (quote set-owner) self)
                            (set! possessions (cons thing possessions))
                            (ask self
                                 (quote say)
                                 (list "I take" (ask thing (quote name)))))
                           (else
                            (ask
                             self
                             (quote say)
                             (list "You cannot take" (ask thing (quote name))))
                            #f))))
                  (lose
                   (lambda (thing)
                     (cond ((eq? self (ask thing (quote owner)))
                            (begin
                             (set! possessions (delq thing possessions))
                             (ask thing (quote set-owner) (quote nobody))
                             (ask self
                                  (quote say)
                                  (list "I lose" (ask thing (quote name)))))
                            (else
                             (ask
                              self
                              (quote say)
                              (list (ask self (quote name))
                                    "does not own"
                                    (ask thing (quote name))))
                             #f)))))
                  (act
                   (lambda ()
                     (let ((direction
                            (pick-random
                             (ask (ask self (quote place)) (quote exits)))))
                       (and direction
                            (ask self (quote go) direction)))))
                  (clock-tick
                   (lambda ()
                     (if (= (random restlessness) 0)
                         (ask self (quote act))
                         (quote no-act))))
                  (go-to
                   (lambda (new-place)
                     (let ((old-place (ask self (quote place))))
                       (cond ((eq? new-place old-place)
                              (ask
                               self
                               (quote say)
                               (list (ask self (quote name))
                                     "is already at"
                                     (ask new-place (quote name))))
                              #f)
                             ((ask mobile-part (quote go-to) new-place)
                              (for-each
                               (lambda (p)
                                 (ask p (quote follow-owner)))
                               possessions)
                              (if (not *turn-off-movement-announcements*)
                                  (ask
                                   self
                                   (quote say)
                                   (list (ask self (quote name))
                                         "moves from"
                                         (ask old-place (quote name))
                                         "to"
                                         (ask new-place (quote name)))))
                              (let ((other-people
                                     (ask self (quote other-people))))
                                (if (and (not *turn-off-greetings*)
                                         (not (null? other-people)))
                                    (ask
                                     self
                                     (quote say)
                                     (cons
                                      "Hi"
                                      (map (lambda (p) (ask p (quote name)))
                                           other-people)))
                                    (quote no-greetings))))
                             (else
                              (ask
                               self
                               (quote say)
                               (list (ask self (quote name))
                                     "can't move to"
                                     (ask new-place (quote name)))))))))
                  (go
                   (lambda (direction)
                     (let ((old-place (ask self (quote place))))
                       (let ((new-place
                              (ask old-place
                                   (quote neighbor-towards)
                                   direction)))
                         (if new-place
                             (ask self (quote go-to) new-place)
                             (ask
                              self
                              (quote say)
                              (list "You cannot go"
                                    direction
                                    "from"
                                    (ask old-place (quote name))))))))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote possessions)) possessions)
                      ((eq? message (quote describe-possessions))
                       describe-possessions)
                      ((eq? message (quote say)) say)
                      ((eq? message (quote have-fit)) have-fit)
                      ((eq? message (quote have-big-fit)) have-big-fit)
                      ((eq? message (quote look-around)) look-around)
                      ((eq? message (quote other-people)) other-people)
                      ((eq? message (quote take)) take)
                      ((eq? message (quote lose)) lose)
                      ((eq? message (quote act)) act)
                      ((eq? message (quote clock-tick)) clock-tick)
                      ((eq? message (quote go-to)) go-to)
                      ((eq? message (quote go)) go)
                      (else (get-method message mobile-part))))))
           (constructor
            (lambda (initial-name initial-location initial-restlessness)
              (set! people (cons self people))
              (set! mobile-part
                    (new-part mobile-class self initial-name initial-location))
              (set! restlessness initial-restlessness)
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value

;Value: things


(define (thing-class)
  (let ((self (quote uninitialized-self))
        (mobile-part (quote uninitialized-part))
        (owner (quote nobody)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote thing-class))
                         (ask mobile-part (quote instance-of?) class-name))))
                  (go-to
                   (lambda (new-place)
                     (or (eq? owner (quote nobody))
                         (ask owner (quote lose) self))
                     (ask mobile-part (quote go-to) new-place)))
                  (follow-owner
                   (lambda ()
                     (if (eq? owner (quote nobody))
                         (reject-move "Thing not owned: "
                                      (ask self (quote name)))
                         (ask mobile-part
                              (quote go-to)
                              (ask owner (quote place))))))
                  (owner (lambda () owner))
                  (owned? (lambda () (not (eq? owner (quote nobody)))))
                  (set-owner
                   (lambda (new-owner)
                     (set! owner new-owner)
                     (quote owner-set))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote go-to)) go-to)
                      ((eq? message (quote follow-owner)) follow-owner)
                      ((eq? message (quote owner)) owner)
                      ((eq? message (quote owned?)) owned?)
                      ((eq? message (quote set-owner)) set-owner)
                      (else (get-method message mobile-part))))))
           (constructor
            (lambda (initial-name initial-location)
              (set! things (cons self things))
              (set! mobile-part
                    (new-part mobile-class self initial-name initial-location))
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value


(define (clock-class)
  (let ((self (quote uninitialized-self))
        (named-part (quote uninitialized-part))
        (time 0)
        (notification-list (quote ())))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote clock-class))
                         (ask named-part (quote instance-of?) class-name))))
                  (tick-tock
                   (lambda ()
                     (newline)
                     (display "---Tick---")
                     (set! time (+ time 1))
                     (for-each
                      (lambda (object)
                        (ask object (quote clock-tick)))
                      notification-list)
                     (quote tick)))
                  (run
                   (lambda (n)
                     (n-times n (lambda () (ask self (quote tick-tock))))
                     (quote done)))
                  (add-to-list
                   (lambda (object)
                     (if (responds-to? object (quote clock-tick))
                         (if (memq object notification-list)
                             (reject-move "Already on clock list: "
                                          (ask object (quote name)))
                             (set! notification-list
                                   (cons object notification-list)))
                         (reject-move "Can not go on clock list: "
                                      (ask object (quote name))))
                     (quote added)))
                  (remove-from-list
                   (lambda (object)
                     (if (memq object notification-list)
                         (set! notification-list
                               (delq object notification-list))
                         (reject-move "not on clock list: "
                                      (ask object (quote name))))
                     (quote removed)))
                  (world-list (lambda () notification-list))
                  (clear-list (lambda () (set! notification-list (quote ()))))
                  (set-time (lambda (new-time) (set! time new-time)))
                  (get-time (lambda () time)))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote tick-tock)) tick-tock)
                      ((eq? message (quote run)) run)
                      ((eq? message (quote add-to-list)) add-to-list)
                      ((eq? message (quote remove-from-list)) remove-from-list)
                      ((eq? message (quote world-list)) world-list)
                      ((eq? message (quote clear-list)) clear-list)
                      ((eq? message (quote set-time)) set-time)
                      ((eq? message (quote get-time)) get-time)
                      (else (get-method message named-part))))))
           (constructor
            (lambda (initial-name)
              (set! named-part (new-part named-class self initial-name))
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value


(define (troll-class)
  (let ((self (quote uninitialized-self))
        (person-part (quote uninitialized-part)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote troll-class))
                         (ask person-part (quote instance-of?) class-name))))
                  (act
                   (lambda ()
                     (let ((others (ask self (quote other-people))))
                       (if (not (null? others))
                           (ask self (quote eat-person) (pick-random others))
                           (ask person-part (quote act))))))
                  (eat-person
                   (lambda (person)
                     (ask
                      self
                      (quote say)
                      (list "Growl.... I'm going to eat you,"
                            (ask person (quote name))))
                     (for-each (lambda (item) (ask person (quote lose) item))
                               (ask person (quote possessions)))
                     (ask person (quote go-to) heaven)
                     (ask person (quote have-big-fit))
                     (ask
                      self
                      (quote say)
                      (list "Chomp chomp. "
                            (ask person (quote name))
                            "tastes yummy!"))
                     (quote *burp*))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote act)) act)
                      ((eq? message (quote eat-person)) eat-person)
                      (else (get-method message person-part))))))
           (constructor
            (lambda (name birthplace threshold)
              (set! person-part
                    (new-part person-class self name birthplace threshold))
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value


(define (rover-class)
  (let ((self (quote uninitialized-self))
        (person-part (quote uninitialized-part))
        (weight-limit 0))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote rover-class))
                         (ask person-part (quote instance-of?) class-name)))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      (else (get-method message person-part))))))
           (constructor
            (lambda (name birthplace threshold initial-weight-limit)
              (set! person-part
                    (new-part person-class self name birthplace threshold))
              (set! weight-limit initial-weight-limit)
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value


(define (transporter-class)
  (let ((self (quote uninitialized-self))
        (place-part (quote uninitialized-part))
        (destination (quote uninitialized-destination)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote transporter-class))
                         (ask person-part (quote instance-of?) class-name)))))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      (else (get-method message place-part))))))
           (constructor
            (lambda (initial-name initial-destination)
              (set! place-part (new-part place-class self initial-name))
              (set! destination initial-destination)
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))
;No value


(define (rock-class)
  (let ((self (quote uninitialized-self))
        (thing-part (quote uninitialized-part))
        (weight (quote uninitialized-weight)))
    (let* ((instance-object
            (let ((instance-of?
                   (lambda (class-name)
                     (or (eq? class-name (quote rock-class))
                         (ask thing-part (quote instance-of?) class-name))))
                  (weight (lambda () weight)))
              (lambda (message)
                (cond ((eq? message (quote instance-of?)) instance-of?)
                      ((eq? message (quote weight)) weight)
                      (else (get-method message thing-part))))))
           (constructor
            (lambda (initial-name initial-location initial-weight)
              (set! thing-part
                    (new-part thing-class self initial-name initial-location))
              (set! weight initial-weight)
              (quote initial-values-set))))
      (define (make-object class-message)
        (cond ((eq? class-message (quote new-part))
               (lambda (new-self . args)
                 (begin (set! self new-self)
                        (apply constructor args)
                        instance-object)))
              ((eq? class-message (quote new))
               (lambda args
                 (begin (set! self instance-object)
                        (apply constructor args)
                        instance-object)))
              (else (fail))))
      make-object)))