;;;                  MASSACHUSETTS INSTITUTE OF TECHNOLOGY
;;;        Department of Electrical Engineering and Computer Science
;;;        6.001---Structure and Interpretation of Computer Programs
;;;                           Fall Semester, 1997
;;;                              Problem Set 7
;;;
;;;                         Code file CLASSES.SCM

;;; -----------------------------------------------------------------------
;;; Basic classes:
;;; An instance of a named class has a name.
;;; An instance of a mobile class has a location that can change.
;;; The person class is an example of a class that responds to clock ticks.
;;; An instance of a place class has neighboring places, and a list of
;;;   things at the place.
;;; An instance of a thing class can be owned.
;;; The clock class is used to create the system clock, which is 
;;;   responsible for simulating the passage of time by sending clock-tick 
;;;   messages to other objects.

;; All classes inherit from this class.
(define-class

  '((class-name named-class)
    
    (parts ())
    
    (slots ((name 'uninitialized-name)))
    
    (constructor
     (lambda (initial-name)
       (set! name initial-name)
       'INITIAL-VALUES-SET))
    
    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (eq? class-name 'NAMED-CLASS)))
      (NAME
       (lambda () name))
      (SAY
       (lambda (list-of-stuff)
	 (newline)
         (if (pair? list-of-stuff)
             (display-message
              (list name "says nothing."))
             (display-message
              (cons name (cons "says" list-of-stuff))))
         'NUF-SAID))))))

; Mobile objects have a location that can change.
(define-class 
  
  '((class-name mobile-class)
    
    (parts (named-part))
    
    (slots ((location 'uninitialized-location)))
    
    (constructor
     (lambda (initial-name initial-location)
       (set! named-part
             (new-part named-class self initial-name))
       (set! location initial-location)
       (ask location 'ADD-THING self)
       'INITIAL-VALUES-SET))
    
    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'MOBILE-CLASS)
             (ask named-part 'INSTANCE-OF? class-name))))
     
      (PLACE
       (lambda () location))
     
      (GO-TO
       (lambda (new-place)
         (ask location 'DEL-THING self)
         (set! location new-place)
         (ask location 'ADD-THING self)
         new-place))))))

(define places '())

(define-class
  
  '((class-name place-class)
    
    (parts (named-part))
    
    (slots ((things '()) (neighbor-map '())))
    
    (constructor
     (lambda (initial-name)
       (set! places (cons self places))
       (set! named-part (new-part named-class self initial-name))
       'INITIAL-VALUES-SET))
    
    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'PLACE-CLASS)
             (ask named-part 'INSTANCE-OF? class-name))))
      
      (THINGS
       (lambda () things))
     
      (NEIGHBORS
       (lambda () (map cadr neighbor-map)))
     
      (EXITS
       (lambda () (map car neighbor-map)))
     
      (NEIGHBOR-TOWARDS
       (lambda (direction)
         (let ((neighbor (assq direction neighbor-map)))
           (and neighbor (cadr neighbor)))))
     
      (ADD-NEIGHBOR
       (lambda (direction new-neighbor)
         (if (ask self 'NEIGHBOR-TOWARDS direction)
             (ask self 'SAY
                  (list (ask self 'NAME)
                        "already has a neighbor to the"
                        direction))
             (begin
               (set! neighbor-map
                     (cons (list direction new-neighbor)
                           neighbor-map))
               'ADDED))))
     
      (HAVE-THING?
       (lambda (thing) (memq thing things)))
     
      ;; The following two methods should not be called by the user.
      ;; They are system-internal methods. See GO-TO instead.
      (ADD-THING
       (lambda (new-thing)
         (or (ask self 'HAVE-THING? new-thing)
             (begin
               (set! things (cons new-thing things))
               #t))))
     
      (DEL-THING
       (lambda (thing)
         (begin
           (if (ask self 'HAVE-THING? thing)
                                        ;DELQ defined in object-system.scm
               (set! things (delq thing things))
               (ask self 'SAY
                    (list (ask thing 'NAME) "is not at" 
                          (ask self 'NAME))))
           #f)))))))


(define people '()) 

;;; Implementation of people
(define-class

  '((class-name person-class)

    (parts (mobile-part))

    (slots 
     ((possessions '())
      (restlessness 'uninitialized-restlessness)))

    (constructor
     (lambda (initial-name initial-location initial-restlessness)
       (set! people (cons self people))
       (set! mobile-part 
             (new-part mobile-class self 
                       initial-name initial-location))
       (set! restlessness initial-restlessness)
       'INITIAL-VALUES-SET))

    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'PERSON-CLASS)
             (ask mobile-part 'INSTANCE-OF? class-name))))

      (POSSESSIONS
       (lambda () possessions))

      (DESCRIBE-POSSESSIONS
       (lambda ()
         (ask self 'SAY
              (cons "I have"
                    (if (null? possessions)
                        '("nothing.")
			(map (lambda (p) (ask p 'NAME))
			     possessions))))))

      (SAY
       (lambda (list-of-stuff)
	 (newline)
         (display-message
          (append (list "At" (ask (ask self 'PLACE) 'NAME)
                        (ask self 'NAME) "says --")
                  (if (null? list-of-stuff)
                      '("Oh, nevermind.")
                      list-of-stuff)))
         'said))

      (HAVE-FIT
       (lambda ()
         (ask self 'SAY '("Yaaaah! I am upset!"))))

      (HAVE-BIG-FIT
       (lambda ()
         (ask self 'SAY 
              '("
                   Grendel, now be still,
                   I kill'd not thee with half so good a will."
                ))
         'GAME-OVER))

      (LOOK-AROUND
       (lambda ()
         (let ((other-things
                (map (lambda (thing) (ask thing 'NAME))
					; DELQ defined in support file
                     (delq self
                           (ask (ask self 'PLACE)
                                'THINGS)))))
           (ask self 'SAY 
                (cons "I see" 
                      (if (null? other-things)
                          '("nothing.")
			  other-things)))
           other-things)))

      (OTHER-PEOPLE
       (lambda ()
         (filter (lambda (object)
                   (if (eq? object self)
                       #f
                       (ask object 'INSTANCE-OF? 'PERSON-CLASS)))
                 (ask (ask self 'PLACE) 'THINGS))))

      (TAKE
       (lambda (thing)
         (cond ((memq thing possessions)
                (ask self 'SAY
                     (list "I already have" 
                           (ask thing 'NAME))))
               ((and (let ((things-at-place 
                            (ask (ask self 'PLACE) 'THINGS)))
                       (memq thing things-at-place))
                     (ask thing 'INSTANCE-OF? 'THING-CLASS))
                (if (ask thing 'OWNED?)
                    (let ((owner (ask thing 'OWNER)))
                      (ask owner 'LOSE thing)
                      (ask owner 'HAVE-FIT))
                    'unowned)

                (ask thing 'SET-OWNER self)
                (set! possessions (cons thing possessions))
                (ask self 'SAY
                     (list "I take" (ask thing 'NAME))))
               (else
		(ask self 'SAY
		     (list "You cannot take" (ask thing 'NAME)))
                #f))))

      (LOSE
       (lambda (thing)
         (cond ((eq? self (ask thing 'OWNER))
					; DELQ defined in object-system.scm
                (begin
                  (set! possessions (delq thing possessions))
                  (ask thing 'SET-OWNER 'nobody)
                  (ask self 'SAY
                       (list "I lose" (ask thing 'NAME)))))
               (else
                (ask self 'SAY
                     (list (ask self 'NAME) 
                           "does not own"
                           (ask thing 'NAME)))
                #f))))

      (ACT
       (lambda ()
         (let ((direction 
                (pick-random (ask (ask self 'PLACE) 'EXITS))))
           (and direction
                (ask self 'GO direction)))))

      (CLOCK-TICK
       (lambda ()
         (if (= (random restlessness) 0)
             (ask self 'ACT)
             'NO-ACT)))

; override the go-to method for mobile objects, so that we can move
; our possessions too.
      (GO-TO
       (lambda (new-place)
         (let ((old-place (ask self 'PLACE)))
           (cond ((eq? new-place old-place)
                  (ask self 'SAY
                       (list (ask self 'NAME)
                             "is already at"
                             (ask new-place 'NAME)))
                  #f)
                 ((ask mobile-part 'GO-TO new-place)
                  (for-each (lambda (p) (ask p 'FOLLOW-OWNER))
                            possessions)
                  (if (not *turn-off-movement-announcements*)
                      (ask self 'SAY
                           (list (ask self 'NAME) 
                                 "moves from"
                                 (ask old-place 'NAME)
                                 "to"
                                 (ask new-place 'NAME))))
                  (let ((other-people (ask self 'OTHER-PEOPLE)))
                    (if (and (not *turn-off-greetings*)
                             (not (null? other-people)))
                        (ask self 'SAY
                             (cons "Hi"
                                   (map (lambda (p) 
                                          (ask p 'NAME))
                                        other-people)))
                        'NO-GREETINGS)))
                 (else
                  (ask self 'SAY
                       (list (ask self 'NAME) 
                             "can't move to"
                             (ask new-place 'NAME))))))))

      (GO
       (lambda (direction)
         (let ((old-place (ask self 'PLACE)))
           (let ((new-place (ask old-place 
                                 'NEIGHBOR-TOWARDS
                                 direction)))
             (if new-place
                 (ask self 'GO-TO new-place)
                 (ask self 'SAY
                      (list "You cannot go"
                            direction
                            "from" 
                            (ask old-place 'NAME))))))))))))

(define things '())

;;; A thing is something that can be owned.
(define-class
  
  '((class-name thing-class)
    
    (parts (mobile-part))
    
    (slots ((owner 'nobody)))
    
    (constructor 
     (lambda (initial-name initial-location)
       (set! things (cons self things))
       (set! mobile-part
             (new-part mobile-class self
                       initial-name initial-location))
       'INITIAL-VALUES-SET))
    
    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'THING-CLASS)
             (ask mobile-part 'INSTANCE-OF? class-name))))
      
      (GO-TO
       (lambda (new-place)
         (or (eq? owner 'nobody)
             (ask owner 'LOSE self))
         (ask mobile-part 'GO-TO new-place)))
      
      (FOLLOW-OWNER
       (lambda ()
         (if (eq? owner 'nobody)
             (reject-move "Thing not owned: " (ask self 'NAME))
             (ask mobile-part 'GO-TO (ask owner 'PLACE)))))
      
      (OWNER
       (lambda () owner))
      
      (OWNED?
       (lambda () (not (eq? owner 'nobody))))
     
      ;; The following method should not be called by the user.
      ;; It is a system-internal method.
      ;; See TAKE and LOSE instead.
      (SET-OWNER
       (lambda (new-owner)
         (set! owner new-owner)
         'OWNER-SET))))))

(define-class
  
  '((class-name clock-class)
    
    (parts (named-part))
    
    (slots 
     ((time 0)
      (notification-list '())))
    
    (constructor
     (lambda (initial-name)
       (set! named-part
             (new-part named-class self initial-name))
       'INITIAL-VALUES-SET))
    
    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'CLOCK-CLASS)
             (ask named-part 'INSTANCE-OF? class-name))))
      
      (TICK-TOCK
       (lambda ()
         (newline)
         (display "---Tick---")
         (set! time (+ time 1))
         (for-each (lambda (object) (ask object 'CLOCK-TICK))
                   notification-list)
         'TICK))
     
      (RUN
       (lambda (n)
         (n-times n (lambda () (ask self 'TICK-TOCK)))
         'DONE))
     
      (ADD-TO-LIST
       (lambda (object)
         (if (responds-to? object 'CLOCK-TICK)
             (if (memq object notification-list)
                 (reject-move "Already on clock list: "
                              (ask object 'NAME))
                 (set! notification-list
                       (cons object notification-list)))
             (reject-move "Can not go on clock list: "
                          (ask object 'NAME)))
         'ADDED
         ))
     
      (REMOVE-FROM-LIST
       (lambda (object)
         (if (memq object notification-list)
             (set! notification-list
                   (delq object notification-list))
             (reject-move "not on clock list: "
                          (ask object 'NAME)))
         'REMOVED))
     
      (WORLD-LIST
       (lambda () notification-list))

      (CLEAR-LIST
       (lambda ()
         (set! notification-list '())))
     
      (SET-TIME
       (lambda (new-time)
         (set! time new-time)))
      
      (GET-TIME
       (lambda () time))))))


;;; -----------------------------------------------------------------------
;;; More interesting classes.

(define-class
  
  '((class-name troll-class)
    
    (parts (person-part))

    (slots ())

    (constructor
     (lambda (name birthplace threshold)
       (set! person-part
             ; *See Problem 6B*
             (new-part person-class self name birthplace threshold)
             )
       'INITIAL-VALUES-SET))

    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'TROLL-CLASS)
             (ask person-part 'INSTANCE-OF? class-name))))

      (ACT
       (lambda ()
         (let ((others (ask self 'OTHER-PEOPLE)))
           (if (not (null? others))
               (ask self 'EAT-PERSON (pick-random others))
               ; *See Problem 6A*
               (ask person-part 'ACT)                       
               ))))

      (EAT-PERSON
       (lambda (person)
         (ask self 'SAY
              (list "Growl.... I'm going to eat you,"
                    (ask person 'NAME)))
         (for-each (lambda (item) (ask person 'LOSE item))
                   (ask person 'POSSESSIONS))
         (ask person 'GO-TO heaven)
         (ask person 'HAVE-BIG-FIT)
         (ask self 'SAY
              (list "Chomp chomp. " (ask person 'NAME)
                    "tastes yummy!"))
         '*burp*))))))


(define-class

  '((class-name rover-class)

    (parts (person-part))

    (slots ((weight-limit 0)))

    (constructor
     (lambda (name birthplace threshold initial-weight-limit)
       (set! person-part
             (new-part person-class self name birthplace threshold))
       (set! weight-limit initial-weight-limit)
       'INITIAL-VALUES-SET))

    (methods
     ((INSTANCE-OF?
       (lambda (class-name)
         (or (eq? class-name 'ROVER-CLASS)
             (ask person-part 'INSTANCE-OF? class-name))))

     ;(ACT
     ; If there are any mars-rocks in the area, pick one
     ; at random and take it.  Then, do what a person
     ; usually does.)
      ))))



(define-class
  '((class-name transporter-class)

    (parts (place-part))

    (slots ((destination 'uninitialized-destination)))

    (constructor
     (lambda (initial-name initial-destination)
       (set! place-part (new-part place-class self initial-name))
       (set! destination initial-destination)
       'INITIAL-VALUES-SET))

    (methods ((INSTANCE-OF?
               (lambda (class-name)
                 (or (eq? class-name 'TRANSPORTER-CLASS)
                     (ask person-part 'INSTANCE-OF? class-name))))

       ;; NOT IMPLEMENTED YET:
       ;; If a rover enters the room, then it should be teleported
       ;; to the transporter's destination.  Make sure that if
       ;; anything else enters the room, everything behaves as
       ;; usual.

               ))))


(define-class
  '((class-name rock-class)

    (parts (thing-part))

    (slots ((weight 'uninitialized-weight)))

    (constructor
     (lambda (initial-name initial-location initial-weight)
       (set! thing-part 
             (new-part thing-class self initial-name 
                       initial-location))
       (set! weight initial-weight)
       'INITIAL-VALUES-SET))

    (methods ((INSTANCE-OF?
               (lambda (class-name)
                 (or (eq? class-name 'ROCK-CLASS)
                     (ask thing-part 'INSTANCE-OF? class-name))))

              (WEIGHT 
               (lambda ()weight))
               ))))


(define (print-state) 
  (newline) 
					;PERSONS
   (for-each
    (lambda (p) 
      (let ((name (ask p 'NAME))
	    (place (ask p 'PLACE))
	    (things (ask p 'POSSESSIONS)))
	(display "person ")
	(display name)
	(if (not (null? things))
	    (begin
	      (display " has ") 
	      (map (lambda (t) (let ((n (ask t 'NAME))) 
				 (display n) (display " ")))
		   things)))
	(display " @ ") 
	(display (ask place 'NAME))
	(newline)))
	     people)
					;THINGS
   (for-each
    (lambda (p) 
      (let ((name (ask p 'NAME))
	    (place (ask p 'PLACE)))
	(begin
	  (display "thing ")
	  (display name) 
	  (if (ask p 'OWNED?)
	      (begin 
		(display " owner ") 
		(display (ask (ask p 'OWNER) 'NAME))))
	  (display " @ ") 
	  (display (ask place 'NAME)) 
	  (newline))))
        things)
					;PLACES
   (for-each
    (lambda (p) 
      (let ((name (ask p 'NAME))
	    (things (ask p 'THINGS)))
	(begin
	  (display "place ") (display name)
	  (if (pair? things)
	      (begin
		(display " : ") 
		(map (lambda (t) (let ((n (ask t 'NAME))) 
				   (display n) (display " ")))
		     things)))
	  (newline))))
       places))
