;;;                  MASSACHVSETTS INSTITVTE OF TECHNOLOGY
;;;        Department of Electrical Engineering and Computer Science
;;;        6.001---Structure and Interpretation of Computer Programs
;;;                           Fall Semester, 1997
;;;                              Problem Set 7
;;;
;;;                       Code file OBJECT-SYSTEM.SCM

;;; -----------------------------------------------------------------------
;;; Procedures for the object oriented adventure game system.

;;THE OBJECT INTERFACE: ASK, NEW and RESPONDS-TO?

; ((object message) arg1 arg2 ...)
(define (ask object message . args)
  (let ((method (get-method message object)))
    (if (fail? method)
        (object-system-error "No method:" message)
        (apply method args))))


; Create and initialize an object.
(define (new class . args)
  (apply ask (cons (class) (cons 'new args))))


(define (responds-to? object message)
  (not (fail? (object message))))


                      ;; OBJECT SYSTEM IMPLEMENTATION


(define (get-method message . objects)
  (if (null? objects)
      (fail)
      (let ((method ((car objects) message)))
        (if (fail? method)
            (apply get-method (cons message (cdr objects)))
            method))))


; Create and initialize a part.
; We must be sure to set self to be the object that this is going to
; be a part of.
(define (new-part class object . args)
  (apply ask
         (cons (class)
               (cons 'new-part
                     (cons object args)))))


(define (get-field field-name association-list)
  (let ((result (assq field-name association-list)))
    (if result
        (cadr result)
        (object-system-error "badly formed class definition:"
                             field-name))))


(define (transform-class-definition class-definition)

  (let ((class-name (get-field 'class-name class-definition))
        (parts (get-field 'parts class-definition))
        (slots (get-field 'slots class-definition))
        (constructor (get-field 'constructor class-definition))
        (methods (get-field 'methods class-definition)))
    (let ((declare-parts 
           (map (lambda (part) `(,part 'uninitialized-part)) parts))
          (instance-object-clauses
           (map (lambda (message-and-method)
                  `((eq? message ',(car message-and-method)) 
                    ,(car message-and-method)))
                methods)))

      `(define (,class-name)

         (let ((self 'uninitialized-self)
               ,@declare-parts 
               ,@slots)

           (let*
               ((instance-object
                 (let ,methods

                   (lambda (message)
                     (cond
                      ,@instance-object-clauses
                      (else (get-method message ,@parts))))))

                (constructor
                 ,constructor))
             (define (make-object class-message)
               (cond
                ((eq? class-message 'new-part)
                 (lambda (new-self . args)
                   (begin
                     (set! self new-self)
                     (apply constructor args)
                     instance-object)))
                ((eq? class-message 'new)
                 (lambda args
                   (begin
                     (set! self instance-object)
                     (apply constructor args)
                     instance-object)))
                (else (fail))))
             make-object))))))
      

;This can be used if you just want to look at the transformed definition.
;See Exercise 11.
;(define (define-class class-definition)
; (pp (transform-class-definition class-definition)))

;THIS IS FOR MITSCHEME ONLY.  FOR OTHER SCHEMES, USE THEIR
;VERSION OF EVAL:
(define (define-class class-definition)
  (eval (transform-class-definition class-definition)
        user-initial-environment))


;;FAIL

(define fail-value (list  "FAIL"))

(define (fail) fail-value)

(define (fail? datum)
  (eq? datum fail-value))


;;; -----------------------------------------------------------------------
;;; Miscellaneous procedures

(define (pick-random lst)
  (if (null? lst)
      false
      (list-ref lst (random (length lst)))))


;;IF NOT MITSCHEME, COMPLETE THIS DEFINITION:
;(define (random n) ...)
;if integer n > 0, returns random integer in [0...(n-1);]


(define (n-times n proc)
  (if (= n 0) 
      'DONE 
      (begin 
        (proc)
        (n-times (- n 1) proc))))


(define (delq item lst)
  (cond ((null? lst) '())
        ((eq? item (car lst)) (delq item (cdr lst)))
        (else (cons (car lst) (delq item (cdr lst))))))


(define (display-message list-of-stuff)
  (define (aux l)
    (if (pair? l)
        (begin
          (display " ")
          (display (car l))
          (aux (cdr l)))))
  (if (pair? list-of-stuff)
      (begin (display (car list-of-stuff))
             (aux (cdr list-of-stuff))))
  'message-displayed)


(define (filter predicate lst)
  (cond ((null? lst) '())
        ((predicate (car lst))
         (cons (car lst) (filter predicate (cdr lst))))
        (else (filter predicate (cdr lst)))))


(define (object-system-error error-name . args)
  (apply error (cons error-name args)))


(define (reject-move error-name . args)
  (display "Move rejected.")
  (newline)
  (display error-name)
  (display-message args)
  'MOVE-REJECTED)


(define *turn-off-movement-announcements* false)
(define *turn-off-greetings* false)

; Create a new symbol

(define (attach-suffix symbol number)
  (string->symbol 
   (string-append 
    (symbol->string symbol) "-" (number->string number))))


;;FOR MITSCHEME ONLY:
;;This is to keep the Scheme printer from going into an infinite loop if you try
;;to print a circular environment structure, such as an object:

(set! *unparser-list-depth-limit* 20)
(set! *unparser-list-breadth-limit* 30)


;IF NOT USING MITSCHEME, load the pset and then execute the following
;expression in the top-level read-eval-print-loop:
;
;(call-with-current-continuation
; (set! object-system-error
;       (lambda (k)
;         (lambda l
;           (k (begin (newline)
;                     (display "OBJECT-SYSTEM-ERROR: ")
;                     (map display l)
;                     (newline)
;                     'error-exit))))))





